@@ -152,6 +152,93 @@ void CallbacksFullFunctionTest(OptimizerType& optimizer,
   REQUIRE(cb.calledStepTaken == calledStepTaken);
 }
 
+template<typename OptimizerType>
+void CallbacksFullMultiobjectiveFunctionTest(OptimizerType& optimizer,
+                                             bool calledEvaluate,
+                                             bool calledGradient,
+                                             bool calledBeginEpoch,
+                                             bool calledEndEpoch,
+                                             bool calledBeginOptimization,
+                                             bool calledEndOptimization,
+                                             bool calledEvaluateConstraint,
+                                             bool calledGradientConstraint,
+                                             bool calledStepTaken)
+{
+  SchafferFunctionN1<arma::mat> SCH;
+
+  typedef decltype(SCH.objectiveA) ObjectiveTypeA;
+  typedef decltype(SCH.objectiveB) ObjectiveTypeB;
+
+  CompleteCallbackTestFunction cb;
+
+  arma::mat coordinates = SCH.GetInitialPoint();
+  std::tuple<ObjectiveTypeA, ObjectiveTypeB> objectives = SCH.GetObjectives();
+
+  optimizer.Optimize(objectives, coordinates, cb);
+
+  REQUIRE(cb.calledEvaluate == calledEvaluate);
+  REQUIRE(cb.calledGradient == calledGradient);
+  REQUIRE(cb.calledBeginEpoch == calledBeginEpoch);
+  REQUIRE(cb.calledEndEpoch == calledEndEpoch);
+  REQUIRE(cb.calledBeginOptimization == calledBeginOptimization);
+  REQUIRE(cb.calledEndOptimization == calledEndOptimization);
+  REQUIRE(cb.calledEvaluateConstraint == calledEvaluateConstraint);
+  REQUIRE(cb.calledGradientConstraint == calledGradientConstraint);
+  REQUIRE(cb.calledStepTaken == calledStepTaken);
+}
+
+template<typename OptimizerType>
+void EarlyStopCallbacksLambdaFunctionTest(OptimizerType& optimizer)
+{
+  arma::mat data, testData, shuffledData;
+  arma::Row<size_t> responses, testResponses, shuffledResponses;
+
+  LogisticRegressionTestData(data, testData, shuffledData,
+      responses, testResponses, shuffledResponses);
+
+  LogisticRegression<> lr(shuffledData, shuffledResponses, 0.5);
+  arma::mat coordinates = lr.GetInitialPoint();
+
+  EarlyStopAtMinLoss cb(
+      [&](const arma::mat& /* coordinates */)
+      {
+        return lr.ComputeAccuracy(testData, testResponses,
+                                  coordinates);
+      });
+
+  optimizer.Optimize(lr, coordinates, cb);
+}
+
+TEST_CASE("EarlyStopAtMinLossLambdaCallbackTest", "[CallbacksTest]")
+{
+  SMORMS3 smorms3;
+  EarlyStopCallbacksLambdaFunctionTest(smorms3);
+}
+
+TEST_CASE("EarlyStopAtMinLossCustomLambdaTest", "[CallbacksTest]")
+{
+  // Use the 50-dimensional Rosenbrock function.
+  GeneralizedRosenbrockFunction f(50);
+  // Start at some really large point.
+  arma::mat coordinates = f.GetInitialPoint();
+  coordinates.fill(100.0);
+
+  EarlyStopAtMinLoss cb(
+      [&](const arma::mat& coordinates)
+      {
+        // Terminate if any coordinate has a value less than 10.
+        double minValue = arma::abs(coordinates).min();
+        return (minValue < 10.0) ?
+          std::numeric_limits<double>::max() : minValue;
+      });
+
+  SMORMS3 smorms3;
+  smorms3.Optimize(f, coordinates, cb);
+
+  // Make sure that we did not get to the optimum.
+  for (size_t i = 0; i < coordinates.n_elem; ++i)
+    REQUIRE(std::abs(coordinates[i]) >= 3.0);
+}
 
 /**
  * Make sure we invoke all callbacks (AdaBound).
@@ -284,6 +371,18 @@ TEST_CASE("KatyushaCallbacksFullFunctionTest", "[CallbacksTest]")
       false, false, true);
 }
 
+/**
+ * Make sure we invoke all callbacks (NSGA2).
+ */
+TEST_CASE("NSGA2CallbacksFullFunctionTest", "[CallbackTest]")
+{
+  arma::vec lowerBound = {-1000};
+  arma::vec upperBound = {1000};
+  NSGA2 optimizer(20, 5000, 0.5, 0.5, 1e-3, 1e-6, lowerBound, upperBound);
+  CallbacksFullMultiobjectiveFunctionTest(optimizer, false, false, false, false,
+      true, true, false, false, true);
+}
+
 /**
  * Make sure we invoke all callbacks (Lookahead).
  */
@@ -550,15 +649,67 @@ TEST_CASE("TimerStopCallbackTest", "[CallbacksTest]")
 {
   SGDTestFunction f;
   arma::mat coordinates = f.GetInitialPoint();
-  
+
   // Instantiate the optimizer with a number of iterations that will take a
   // long time to finish.
   Adam opt(0.5, 2, 0.7, 0.999, 1e-8, 2000000000, -100, false);
   arma::wall_clock timer;
-  
+
   timer.tic();
   // The optimization process should return in one second.
   opt.Optimize(f, coordinates, TimerStop(0.5));
   // Add some time to account for the function to return.
   REQUIRE(timer.toc() < 2);
 }
+
+/**
+ * Make sure the ProgressBar callback will show the progress on the specified
+ * output stream if the MaxIterations parameter of the optimizer is 0.
+ */
+TEST_CASE("ProgressBarCallbackNoMaxIterationsTest", "[CallbacksTest]")
+{
+  SGDTestFunction f;
+  arma::mat coordinates = f.GetInitialPoint();
+
+  StandardSGD s(0.0003, 1, 0, DBL_MAX, true);
+
+  std::stringstream stream;
+  s.Optimize(f, coordinates, ProgressBar(10, stream));
+
+  REQUIRE(stream.str().length() > 0);
+}
+
+/**
+ * Make sure the ProgressBar callback will show the progress on the specified
+ * output stream with the correct epoch number if the MaxIterations parameter
+ * of the optimizer is 0.
+ */
+TEST_CASE("ProgressBarCallbackNoMaxIterationsEpochTest", "[CallbacksTest]")
+{
+  SGDTestFunction f;
+  arma::mat coordinates = f.GetInitialPoint();
+
+  StandardSGD s(0.0003, 1, 0, DBL_MAX, true);
+
+  std::stringstream stream;
+  s.Optimize(f, coordinates, ProgressBar(10, stream));
+  REQUIRE(stream.str().find("Epoch 1") != std::string::npos);
+  REQUIRE(stream.str().find("Epoch 1/") == std::string::npos);
+}
+
+/**
+ * Make sure the ProgressBar callback will show the progress on the specified
+ * output stream with the correct epoch number if the MaxIterations parameter
+ * of the optimizer is not equal to 0.
+ */
+TEST_CASE("ProgressBarCallbackEpochTest", "[CallbacksTest]")
+{
+  SGDTestFunction f;
+  arma::mat coordinates = f.GetInitialPoint();
+
+  StandardSGD s(0.0003, 1, 1, 1e-9, true);
+
+  std::stringstream stream;
+  s.Optimize(f, coordinates, ProgressBar(10, stream));
+  REQUIRE(stream.str().find("Epoch 1/1") != std::string::npos);
+}
